# !/usr/bin/python3
"""
pdb_eda single structure analysis mode command-line interface
  Analyzes a single pdb entry.

Usage:
    pdb_eda single -h | --help
    pdb_eda single <pdbid> <out-file> map (--density | --diff-density)
    pdb_eda single <pdbid> <out-file> cloud (--atom | --residue | --domain) [--out-format=<format>] [--params=<params-file>] [--include-pdbid]
    pdb_eda single <pdbid> <out-file> density (--atom | --residue | --symmetry-atom) [--type=<type>] [--radius=<radius>] [--num-sd=<num-sd>] [--out-format=<format>] [--params=<params-file>] [--include-pdbid] [--atom-mask=<mask-file>] [--optimized-radii]
    pdb_eda single <pdbid> <out-file> difference (--atom | --residue | --symmetry-atom) [--type=<type>] [--radius=<radius>] [--num-sd=<num-sd>] [--out-format=<format>] [--params=<params-file>] [--include-pdbid] [--atom-mask=<mask-file>]
    pdb_eda single <pdbid> <out-file> blob [--green] [--red] [--num-sd=<num-sd>] [--out-format=<format>] [--params=<params-file>] [--include-pdbid]
    pdb_eda single <pdbid> <out-file> blob --blue [--num-sd=<num-sd>] [--out-format=<format>] [--params=<params-file>] [--include-pdbid]
    pdb_eda single <pdbid> <out-file> statistics [--out-format=<format>] (--atom | --residue) [--include-pdbid] [--print-validation]

Options:
    -h, --help                      Show this screen.
    <pdbid>                         The PDB ID to download and analyze.
    <out-file>                      Output filename. "-" will write to standard output.
    --params=<params-file>          Overriding parameters file that includes radii, slopes, etc. [default: ]
    --include-pdbid                 Include PDB ID at the beginning of each result.
    --density                       Output the density 2Fo-Fc map in jsonpickle format.
    --diff-density                  Output the difference density Fo-Fc map in jsonpickle format.
    --atom                          Calculate results for each atom.
    --residue                       Calculate results for each residue.
    --symmetry-atom                 Calculate results for each symmetry atom.
    --domain                        Calculate results for each domain.
    --green                         Calculate green (positive) difference map blobs.
    --red                           Calculate red (negative) difference map blobs.
    --blue                          Calculate blue (positive) density map blobs. Default option if red/green not selected.  However, this option uses a LOT OF MEMORY at 1.5sd.
    --radius=<radius>               Radius (in angstroms) around atom or residue to calculate significant discrepancy. [default: 3.5]
    --num-sd=<num-sd>               Number of standard deviation units to use as a significant discrepancy cutoff. Default is 3.0 for atom, residue, green, and red.  Default is 1.5 for blue.
    --atom-mask=<mask-file>         JSON file with dictionary of residue type to list of atom types to use in calculating residue-specific regional electron density and discrepancies.
    --optimized-radii               Use optimized atom radii when available.
    --type=<type>                   Residue type or atom type to filter by.
    --out-format=<format>           Output file format, available formats: csv, json [default: json].
    --print-validation              Print comparison of median absolute values below 1 sigma for Fo and Fc maps, which should be very similar.

Submodes:
    map                             Output the electron density map in jsonpickle format.
    cloud                           Output the 2Fo-Fc density clouds based on atom-specific radii. Gives better estimate of atom/residue/domain-specific electron density, but only for optimized atom radii.
    density                         Output the regional 2Fo-Fc density. (**User must select the desired radius, since the 3.5 angstrom default is likely not appropriate).
    difference                      Output the regional Fo-Fc density discrepancy.
    blob                            Output either 2Fo-Fc (blue) density blobs or Fo-Fc (green or red) difference density blobs.
    statistics                      Output statistics at the atom, residue, or whole electron density map level.
"""

import docopt
import sys
import json
import jsonpickle
import numpy

from . import densityAnalysis
from . import __version__

def main():
    args = docopt.docopt(__doc__, version=__version__)
    if args["--help"]:
        print(__doc__)
        exit(0)

    args["--radius"] = float(args["--radius"])

    if args["--num-sd"] == None:
        args["--num-sd"] = 3.0 if args["--green"] or args["--red"] or args["difference"] else 1.5
    args["--num-sd"] = float(args["--num-sd"])

    if args["--params"]:
        try:
            with open(args["--params"], 'r') as paramsFile:
                params = json.load(paramsFile)
            densityAnalysis.setGlobals(params)
        except:
            raise RuntimeError(str("Error: params file \"") + args["--params"] + "\" does not exist or is not parsable.")

    atom_mask = None
    if args["--atom-mask"]:
        try:
            with open(args["--atom-mask"], 'r') as atomMaskFile:
                atom_mask = json.load(atomMaskFile)
        except:
            raise RuntimeError(str("Error: atom mask file \"") + args["--atom-mask"] + "\" does not exist or is not parsable.")


    analyzer = densityAnalysis.fromPDBid(args["<pdbid>"])
    if not analyzer:
        raise RuntimeError("Error: Unable to parse or download PDB entry or associated ccp4 file.")

    jsonType = "json"
    if args["--density"]:
        jsonType = "jsonpickle"
        result = analyzer.densityObj
    elif args["--diff-density"]:
        jsonType = "jsonpickle"
        result = analyzer.diffDensityObj
    elif args["cloud"]:
        analyzer.aggregateCloud()
        if args["--atom"]:
            headerList = list(map(str, list(analyzer.atomCloudDescriptions.dtype.names) + ['density_electron_ratio']))
            result = [[numpyConverter(element) for element in item] + [analyzer.densityElectronRatio] for item in analyzer.atomCloudDescriptions]
        elif args["--residue"]:
            headerList = densityAnalysis.DensityAnalysis.residueCloudHeader + ['density_electron_ratio']
            result = [list(item) + [analyzer.densityElectronRatio] for item in analyzer.residueCloudDescriptions]
        elif args["--domain"]:
            headerList = densityAnalysis.DensityAnalysis.domainCloudHeader + ['density_electron_ratio']
            result = [list(item) + [analyzer.densityElectronRatio] for item in analyzer.domainCloudDescriptions]
    elif args["density"]:
        if args["--atom"]:
            headerList = densityAnalysis.DensityAnalysis.atomRegionDensityHeader
            result = analyzer.calculateAtomRegionDensity(args["--radius"], args["--num-sd"], args["--type"], args["--optimized-radii"])
        elif args["--residue"]:
            headerList = densityAnalysis.DensityAnalysis.residueRegionDensityHeader
            result = analyzer.calculateResidueRegionDensity(args["--radius"], args["--num-sd"], args["--type"], atom_mask, args["--optimized-radii"])
        elif args["--symmetry-atom"]:
            headerList = densityAnalysis.DensityAnalysis.symmetryAtomRegionDensityHeader
            result = analyzer.calculateSymmetryAtomRegionDensity(args["--radius"], args["--num-sd"], args["--type"], args["--optimized-radii"])
            for atomInfo in result:
                atomInfo[4] = [val for val in atomInfo[4]]
                atomInfo[5] = [float(val) for val in atomInfo[5]]
    elif args["difference"]:
        if args["--atom"]:
            headerList = densityAnalysis.DensityAnalysis.atomRegionDiscrepancyHeader
            result = analyzer.calculateAtomRegionDiscrepancies(args["--radius"], args["--num-sd"], args["--type"])
        elif args["--residue"]:
            headerList = densityAnalysis.DensityAnalysis.residueRegionDiscrepancyHeader
            result = analyzer.calculateResidueRegionDiscrepancies(args["--radius"], args["--num-sd"], args["--type"], atom_mask)
        elif args["--symmetry-atom"]:
            headerList = densityAnalysis.DensityAnalysis.symmetryAtomRegionDiscrepancyHeader
            result = analyzer.calculateSymmetryAtomRegionDiscrepancies(args["--radius"], args["--num-sd"], args["--type"])
            for atomInfo in result:
                atomInfo[4] = [val for val in atomInfo[4]]
                atomInfo[5] = [float(val) for val in atomInfo[5]]
    elif args['blob']:
        headerList = densityAnalysis.DensityAnalysis.blobStatisticsHeader
        result = []
        if args["--green"]:
            greenBlobList = analyzer.diffDensityObj.createFullBlobList(analyzer.diffDensityObj.meanDensity + args["--num-sd"] * analyzer.diffDensityObj.stdDensity)
            result.extend(analyzer.calculateAtomSpecificBlobStatistics(greenBlobList))
        if args["--red"]:
            redBlobList = analyzer.diffDensityObj.createFullBlobList(-1 * (analyzer.diffDensityObj.meanDensity + args["--num-sd"] * analyzer.diffDensityObj.stdDensity))
            result.extend(analyzer.calculateAtomSpecificBlobStatistics(redBlobList))
        if not args["--green"] and not args["--red"]: # args["--blue"] by default
            blueBlobList = analyzer.densityObj.createFullBlobList(analyzer.densityObj.meanDensity + args["--num-sd"] * analyzer.densityObj.stdDensity)
            result.extend(analyzer.calculateAtomSpecificBlobStatistics(blueBlobList))
        for blobInfo in result:
            blobInfo[9]  = [ val for val in blobInfo[9] ]
            blobInfo[10] = [ float(val) for val in blobInfo[10] ]
            blobInfo[11] = [ float(val) for val in blobInfo[11] ]
    elif args["statistics"]:
        if args["--print-validation"]:
            (medianAbsFo,medianAbsFc) = analyzer.medianAbsFoFc()
            print("Median abs Fo(<1sd):", medianAbsFo, "Median abs Fc(<1sd):", medianAbsFc,"Relative Difference:",(medianAbsFo - medianAbsFc)/max(medianAbsFo,medianAbsFc))

        if args["--residue"]:
            headerList = analyzer.residueMetricsHeaderList
            result = analyzer.residueMetrics()
        elif args["--atom"]:
            headerList = analyzer.atomMetricsHeaderList
            result = analyzer.atomMetrics()
            for atomInfo in result:
                atomInfo[4] = [x for x in atomInfo[4]]
                atomInfo[5] = [float(x) for x in atomInfo[5]]

    if args["--include-pdbid"]:
        headerList = ["pdbid"] + headerList
        result = [ [analyzer.pdbid] + row for row in result ]

    with open(args["<out-file>"], 'w') if args["<out-file>"] != "-" else sys.stdout as outFile:
        if args["--out-format"] == 'csv':
            csvResults = [','.join(map(str, row)) for row in [headerList] + result]
            print(*csvResults, sep='\n', file=outFile)
        elif jsonType == "jsonpickle":
            jsonText = jsonpickle.encode(result)
            outFile.write(jsonText)
        else:
            jsonResults = [ dict(zip(headerList, row)) for row in result ]
            print(json.dumps(jsonResults, indent=2, sort_keys=True), file=outFile)

def numpyConverter(obj):
    """Converts numpy objects to standard Python types, otherwise just return the object.

    :param obj:
    :type obj: :py:class:`object`

    :return: object
    :rtype: :py:class:`int`, :py:class:`float`, :py:class:`list`, :py:class:`object`
    """
    if isinstance(obj, numpy.integer):
        return int(obj)
    elif isinstance(obj, numpy.floating):
        return float(obj)
    elif isinstance(obj, numpy.ndarray):
        return [numpyConverter(item) for item in obj]
    else:
        return obj



# Compile using:
# $ python3 cutils_setup.py build_ext --inplace

from setuptools import setup, Extension
import Cython.Build

setup(
  name = 'cutils',
  ext_modules=[
    Extension('cutils',
              sources=['cutils.pyx'],
              extra_compile_args=['-O3'],
              language='c++')
    ],
  cmdclass = {'build_ext': Cython.Build.build_ext}
)



#!/usr/bin/python3
"""
pdb_eda (crystal) contacts analysis mode command-line interface
	Analyzes atoms in a PDB entry for the closest crystal contacts.
	This mode requires the pymol package and associated python library to be installed.

Usage:
	pdb_eda contacts -h | --help
	pdb_eda contacts <pdbid> <out-file> [--distance=<cutoff>] [--symmetry-atoms] [--include-pdbid] [--out-format=<format>]

Options:
    -h, --help             Show this screen.
    <pdbid>                The PDB ID to download and analyze.
    <out-file>             Output filename. "-" will write to standard output.
    --distance=<cutoff>    Distance cutoff in angstroms for detecting crystal contacts [default: 5.0].
    --symmetry-atoms       Calculate crystal contacts to symmetry atoms too.
    --include-pdbid        Include PDB ID at the beginning of each result.
    --out-format=<format>  Output file format, available formats: csv, json [default: json].
"""

import scipy.spatial.distance
import numpy as np
import sys
import os
import docopt
import json

from . import densityAnalysis
from . import __version__

try:
	import pymol
	pymolImported = True
except ImportError:
	pymolImported = False

def main():
	if not pymolImported:
		print("Error: pymol module did not load.")
		print(__doc__)
		exit(0)

	args = docopt.docopt(__doc__, version=__version__)

	if args["--help"]:
		print(__doc__)
		exit(0)

	args["--distance"] = float(args["--distance"])
	args["<pdbid>"] = args["<pdbid>"].lower()

	analyzer = densityAnalysis.fromPDBid(args["<pdbid>"], mmcif=True)
	if not analyzer:
		raise RuntimeError("Error: Unable to parse or download PDB entry or associated ccp4 file.")

	mmcif_file = densityAnalysis.pdbfolder + args["<pdbid>"] + '.cif.gz'

	crystalNeighborCoords = simulateCrystalNeighborCoordinates(mmcif_file, args["--distance"])

	if args["--symmetry-atoms"]:
		atoms = analyzer.symmetryAtoms
		contacts = findCoordContacts(analyzer.symmetryAtomCoords, crystalNeighborCoords, args["--distance"])
	else:
		atoms = list(analyzer.biopdbObj.get_atoms())
		contacts = findCoordContacts(np.asarray([atom.coord for atom in atoms]), crystalNeighborCoords, args["--distance"])

	headerList = ['model', 'chain', 'residue_number', 'residue_name', "atom_name", "occupancy", "symmetry", "xyz", "crystal_contact_distance"]
	result = []
	for index,contactDistance in contacts:
		atom = atoms[index]
		result.append([atom.parent.parent.parent.id, atom.parent.parent.id, atom.parent.id[1], atom.parent.resname, atom.name, atom.get_occupancy(),
					   [x for x in atom.symmetry] if args["--symmetry-atoms"] else [0,0,0,0], [float(c) for c in atom.coord], contactDistance])

	if args["--include-pdbid"]:
		headerList = ["pdbid"] + headerList
		result = [[analyzer.pdbid] + row for row in result]

	with open(args["<out-file>"], 'w') if args["<out-file>"] != "-" else sys.stdout as outFile:
		if args["--out-format"] == 'csv':
			csvResults = [','.join(map(str, row)) for row in [headerList] + result]
			print(*csvResults, sep='\n', file=outFile)
		else:
			jsonResults = [dict(zip(headerList, row)) for row in result]
			print(json.dumps(jsonResults, indent=2, sort_keys=True), file=outFile)


def findCoordContacts(coordList1, coordList2, distanceCutoff=5.0):
	"""Find contacts in coordList1 to coordList2 at the given distance cutoff.

	:param coordList1: list of coordinates.
	:type coordList1: :py:class:`list`
	:param coordList2: list of coordinates.
	:type coordList2: py:class:`list`
	:param distanceCutoff: distance cutoff., defaults to 5.0
	:type distanceCutoff: :py:class:`float`

	:return: contactList of index,minDistance tuples.
	:rtype: :py:class:`list`
	"""
	distances = scipy.spatial.distance.cdist(coordList1, coordList2)
	return [(index,minDistance) for index,minDistance in enumerate(np.min(distances[x]) for x in range(len(coordList1))) if minDistance <= distanceCutoff]


def simulateCrystalNeighborCoordinates(filename, distanceCutoff=5.0):
	"""RETURN a list of atom coordinates of the simulated crystal environment surrounding the X-Ray
	Diffraction asymmetric unit (excluding heteroatoms). Requires a file path instead of a
	structure because the bulk of this is handled by Pymol.
	NOTE: This will only work with PDB structures resolved with X-RAY DIFFRACTION.

	:param filename:
	:type filename: :py:class:`str`
	:param distanceCutoff: distance cutoff., defaults to 5.0
	:type distanceCutoff: :py:class:`float`

	:return: coordList
	:rtype: :py:class:`list`
	"""
	# Launch pymol.
	pymol.pymol_argv = ['pymol', '-qc']
	pymol.finish_launching()

	# Set absolute path and name of PDB entry.
	spath = os.path.abspath(filename)
	sname = spath.split('/')[-1].split('.')[0]
	asym_unit = "asym_unit"

	# Load Structure.
	pymol.cmd.load(spath)
	pymol.cmd.disable("all")
	pymol.cmd.enable(sname)
	pymol.cmd.create(asym_unit, "polymer")

	# Generate local crystal environment using symexp; delete original structures.
	pymol.cmd.symexp("neighbor", asym_unit, asym_unit, distanceCutoff)
	pymol.cmd.delete(sname)
	pymol.cmd.delete(asym_unit)

	# Return atom coordinates of the neighboring structures.
	myspace = {"coordinates": []}
	pymol.cmd.iterate_state(1, "all", "coordinates.append([x,y,z])", space=myspace)

	pymol.cmd.reinitialize()
	return myspace["coordinates"]




"""
pdb_eda multiple structure analysis mode command-line interface
  Analyzes multiple pdb entries.

Usage:
    pdb_eda multiple -h | --help
    pdb_eda multiple <pdbid-file> <out-result-file> [--params=<params-file>] [--out-format=<format>] [--testing] [--time-out=<seconds>] [--silent]
    pdb_eda multiple <in-result-file> <out-pdbid-file> --filter [--out-format=<format>] [--max-resolution=<max-resolution>] [--min-resolution=<min-resolution>] [--min-atoms=<min-atoms>] [--min-residues=<min-residues>] [--min-elements=<min-elements>]
    pdb_eda multiple <pdbid-file> --reload
    pdb_eda multiple <pdbid-file> <out-dir> --single-mode=<quoted-single-mode-options> [--time-out=<seconds>] [--testing] [--silent] [--skip]
    pdb_eda multiple <pdbid-file> <out-dir> --contacts-mode=<quoted-contacts-mode-options> [--time-out=<seconds>] [--testing] [--silent] [--skip] [--safe]

Options:
    -h, --help                                      Show this screen.
    <out-result-file>                               Output filename. "-" will write to standard output.
    <in-result-file>                                Input filename. "-" will read from standard input.  Format must be JSON.
    <pdbid-file>                                    Input filename that contains the pdb ids. "-" will read from standard input.
    <out-pdbid-file>                                Output filename that contains the pdb ids. "-" will write to standard output.
    --params=<params-file>                          Overriding parameters file that includes radii, slopes, etc. [default: ]
    --out-format=<format>                           Output file format, available formats: csv, json [default: json].
    --time-out=<seconds>                            Set a maximum time to try to analyze any single pdb entry. [default: 0]
    --testing                                       Run only a single process for testing purposes.
    --silent                                        Do not print error message to stderr.
    --single-mode=<quoted-single-mode-options>      Run single structure analysis mode on a set of PDB entries.
    --contacts-mode=<quoted-contacts-mode-options>  Run (crystal) contacts analysis mode on a set of PDB entries.
    --skip                                          Skip pdbids that where previously analyzed.
    --safe                                          Run as a separate subprocess, since the pymol module is not always stable for multiple executions.
    --filter                                        Filter pdbids based on criteria.
    --max-resolution=<max-resolution>               Maximum x-ray crystallographic resolution to allow. [default: 3.5]
    --min-resolution=<min-resolution>               Minimum x-ray crystallographic resolution to allow. [default: 0]
    --min-atoms=<min-atoms>                         Minimum number of atoms analyzed (numAtomsAnalyzed) required. [default: 300]
    --min-residues=<min-residues>                   Minimum number of residues required of the given residue types separated by commas. [default: 0]
    --min-elements=<min-elements>                   Minimum number of atoms required of the given elements separated by commas. [default: 0]
    --reload                                        Test that each pdbid loads and try to download again any pdbid that fail.

List of all wwPDB entry IDs along with entry type.
ftp://ftp.wwpdb.org/pub/pdb/derived_data/pdb_entry_type.txt
"""

import os
import gc
import time
import sys

import json
import csv
import multiprocessing
import subprocess
import signal
import collections
import docopt

from . import densityAnalysis
from . import __version__
from . import singleStructure
from . import crystalContacts
from . import fileUtils

globalArgs = {}

def main():
    global globalArgs
    globalArgs = docopt.docopt(__doc__, version=__version__)
    if globalArgs["--help"]:
        print(__doc__)
        exit(0)
    globalArgs["--time-out"] = int(globalArgs["--time-out"])

    if globalArgs["--params"]:
        try:
            with open(globalArgs["--params"], 'r') as paramsFile:
                params = json.load(paramsFile)
            densityAnalysis.setGlobals(params)
        except:
            raise RuntimeError(str("Error: params file \"") + globalArgs["--params"] + "\" does not exist or is not parsable.")

    if globalArgs["--filter"]:
        globalArgs["--max-resolution"] = float(globalArgs["--max-resolution"])
        globalArgs["--min-resolution"] = float(globalArgs["--min-resolution"])
        globalArgs["--min-atoms"] = int(globalArgs["--min-atoms"])

        if "," in globalArgs["--min-residues"]:
            allowedResidueTypes = globalArgs["--min-residues"].split(",")
            globalArgs["--min-residues"] = float(allowedResidueTypes.pop(0))
            allowedResidueTypes = set(allowedResidueTypes)
        else:
            globalArgs["--min-residues"] = float(globalArgs["--min-residues"])
            allowedResidueTypes = set()

        if "," in globalArgs["--min-elements"]:
            allowedElements = globalArgs["--min-elements"].split(",")
            globalArgs["--min-elements"] = float(allowedElements.pop(0))
            allowedElements = set(allowedElements)
        else:
            globalArgs["--min-elements"] = float(globalArgs["--min-elements"])
            allowedElements = set()

        try:
            with open(globalArgs["<in-result-file>"], "r") if globalArgs["<in-result-file>"] != "-" else sys.stdin as jsonFile:
                pdbResults = json.load(jsonFile)
        except:
            raise RuntimeError(str("Error: PDB results file \"") + globalArgs["<in-result-file>"] + "\" does not exist or is not parsable.")

        pdbids = [pdbid for pdbid in pdbResults if
                  pdbResults[pdbid]["stats"]["num_atoms_analyzed"] >= globalArgs["--min-atoms"] and
                  float(pdbResults[pdbid]["stats"]["resolution"]) >= globalArgs["--min-resolution"] and
                  float(pdbResults[pdbid]["stats"]["resolution"]) <= globalArgs["--max-resolution"] and
                  sum(count for (residueType, count) in pdbResults[pdbid]["properties"]["residue_counts"].items() if not allowedResidueTypes or residueType in allowedResidueTypes) >= globalArgs["--min-residues"] and
                  sum(count for (element, count) in pdbResults[pdbid]["properties"]["element_counts"].items() if not allowedElements or element in allowedElements) >= globalArgs["--min-elements"]
                  ]

        if globalArgs["--out-format"] == "json":
            with open(globalArgs['<out-pdbid-file>'], "w") if globalArgs["<out-pdbid-file>"] != "-" else sys.stdout as jsonFile:
                print(json.dumps(pdbids, indent=2, sort_keys=True), file=jsonFile)
        else:
            with open(globalArgs['<out-pdbid-file>'], "w") if globalArgs["<out-pdbid-file>"] != "-" else sys.stdout as txtFile:
                print("\n".join(pdbids), file=txtFile)
    elif globalArgs["--reload"]:
        try:
            pdbids = []
            with open(globalArgs["<pdbid-file>"], "r") if globalArgs["<pdbid-file>"] != "-" else sys.stdin as textFile:
                for pdbid in textFile:
                    pdbids.append(pdbid[0:4])
        except:
            raise RuntimeError(str("Error: PDB IDs file \"") + globalArgs["<pdbid-file>"] + "\" does not exist or is not parsable.")

        badPDBids = [pdbid for pdbid in pdbids if not testPDBIDLoad(pdbid)]
        for pdbid in badPDBids:
            densityAnalysis.cleanPDBid(pdbid)
        badPDBids = [pdbid for pdbid in badPDBids if not testPDBIDLoad(pdbid)]
        for pdbid in badPDBids:
            densityAnalysis.cleanPDBid(pdbid)
        if badPDBids:
            print("Bad PDBids:",",".join(badPDBids))
    else:
        try:
            pdbids = []
            with open(globalArgs["<pdbid-file>"], "r") if globalArgs["<pdbid-file>"] != "-" else sys.stdin as textFile:
                for pdbid in textFile:
                    pdbids.append(pdbid[0:4])
        except:
            raise RuntimeError(str("Error: PDB IDs file \"") + globalArgs["<pdbid-file>"] + "\" does not exist or is not parsable.")

        if globalArgs["--single-mode"]:
            processFunction = singleModeFunction
            if not os.path.isdir(globalArgs["<out-dir>"]):
                if not os.path.isfile(globalArgs["<out-dir>"]):
                    os.mkdir(globalArgs["<out-dir>"])
                else:
                    raise RuntimeError(str("Error: Output directory \"") + globalArgs["<out-dir>"] + "\" is a file.")
        elif globalArgs["--contacts-mode"]:
            processFunction = contactsModeFunction if not globalArgs["--safe"] else contactsModeFunctionSafe
            if not os.path.isdir(globalArgs["<out-dir>"]):
                if not os.path.isfile(globalArgs["<out-dir>"]):
                    os.mkdir(globalArgs["<out-dir>"])
                else:
                    raise RuntimeError(str("Error: Output directory \"") + globalArgs["<out-dir>"] + "\" is a file.")
        else:
            processFunction = multipleModeFunction

        if globalArgs["--skip"]:
            pdbids = [ pdbid for pdbid in pdbids if not os.path.isfile(globalArgs["<out-dir>"] + "/" + pdbid + ".result") ]

        if globalArgs["--testing"]:
            results = [ processFunction(pdbid) for pdbid in pdbids ]
        else:
            with multiprocessing.Pool() as pool:
                results = pool.map(processFunction, pdbids, chunksize=1)

        if not globalArgs["--single-mode"] and not globalArgs["--contacts-mode"]: # skip generating results if running in another mode or submode.
            fullResults = {}
            for resultFilename in results:
                if resultFilename:
                    try:
                        with open(resultFilename, 'r') as jsonFile:
                            result = json.load(jsonFile)
                            fullResults[result["pdbid"]] = result
                        os.remove(resultFilename)
                    except:
                        pass

            if globalArgs["--out-format"] == 'csv' or globalArgs["--out-format"] == 'txt':
                statsHeaders = ['density_electron_ratio', 'voxel_volume', 'f000', 'num_voxels_aggregated', 'total_aggregated_electrons', 'density_mean', 'diff_density_mean', 'resolution', 'space_group',
                                'num_atoms_analyzed', 'num_residue_clouds_analyzed', 'num_domain_clouds_analyzed', "atom_overlap_completeness"]
                with open(globalArgs['<out-result-file>'], "w", newline='') if globalArgs["<out-result-file>"] != "-" else sys.stdout as csvFile:
                    writer = csv.writer(csvFile)
                    writer.writerow(['pdbid'] + statsHeaders + sorted(densityAnalysis.paramsGlobal["radii"]))
                    for result in fullResults.values():
                        stats = [result["stats"][header] for header in statsHeaders]
                        diffs = [result["diffs"][atomType] for atomType in sorted(densityAnalysis.paramsGlobal["radii"])]
                        writer.writerow([result['pdbid']] + stats + diffs)
            else:
                with open(globalArgs['<out-result-file>'], "w") if globalArgs["<out-result-file>"] != "-" else sys.stdout as jsonFile:
                    print(json.dumps(fullResults, indent=2, sort_keys=True), file=jsonFile)


def singleModeFunction(pdbid):
    """Execute pdb_eda single mode on a given pdbid.

    :param pdbid:
    :type pdbid: :py:class:`str`

    :return: 0
    """
    singleModeCommandLine = "pdb_eda single " + pdbid + " " + globalArgs["<out-dir>"] + "/" + pdbid + ".result " + globalArgs["--single-mode"]
    sys.argv = singleModeCommandLine.split()
    if globalArgs["--time-out"]:
        try:
            with timeout(seconds=globalArgs["--time-out"]):
                singleStructure.main()
        except Exception as exception:
            if not globalArgs["--silent"]:
                print(pdbid, exception, file=sys.stderr)
    elif globalArgs["--testing"]:
        singleStructure.main()
    else:
        try:
            singleStructure.main()
        except Exception as exception:
            if not globalArgs["--silent"]:
                print(pdbid, exception, file=sys.stderr)

    gc.collect()
    return 0


def contactsModeFunction(pdbid):
    """Execute pdb_eda contacts mode on a given pdbid.

    :param pdbid:
    :type pdbid: :py:class:`str`

    :return: 0
    """
    contactsModeCommandLine = "pdb_eda contacts " + pdbid + " " + globalArgs["<out-dir>"] + "/" + pdbid + ".result " + globalArgs["--contacts-mode"]
    sys.argv = contactsModeCommandLine.split()
    if globalArgs["--time-out"]:
        try:
            with timeout(seconds=globalArgs["--time-out"]):
                crystalContacts.main()
        except Exception as exception:
            if not globalArgs["--silent"]:
                print(pdbid, exception, file=sys.stderr)
    elif globalArgs["--testing"]:
        crystalContacts.main()
    else:
        try:
            crystalContacts.main()
        except Exception as exception:
            if not globalArgs["--silent"]:
                print(pdbid, exception, file=sys.stderr)

    gc.collect()
    return 0


def contactsModeFunctionSafe(pdbid):
    """Execute pdb_eda contacts mode on a given pdbid as a separate subprocess.

    :param pdbid:
    :type pdbid: :py:class:`str`

    :return: 0
    """
    contactsModeCommandLine = "pdb_eda contacts " + pdbid + " " + globalArgs["<out-dir>"] + "/" + pdbid + ".result " + globalArgs["--contacts-mode"]
    args = contactsModeCommandLine.split()
    if globalArgs["--time-out"]:
        try:
            with timeout(seconds=globalArgs["--time-out"]):
                subprocess.run(args)
        except Exception as exception:
            if not globalArgs["--silent"]:
                print(pdbid, exception, file=sys.stderr)
    elif globalArgs["--testing"]:
        subprocess.run(args)
    else:
        try:
            subprocess.run(args)
        except Exception as exception:
            if not globalArgs["--silent"]:
                print(pdbid, exception, file=sys.stderr)

    gc.collect()
    return 0



def multipleModeFunction(pdbid):
    """Analyze a single pdbid.

    :param pdbid:
    :type pdbid: py:class:`str`

    :return: resultFilename or 0
    :rtype: :py:class:`str`, :py:class:`int`
    """
    if globalArgs["--time-out"]:
        try:
            with timeout(seconds=globalArgs["--time-out"]):
                return analyzePDBID(pdbid)
        except Exception as exception:
            if not globalArgs["--silent"]:
                print(pdbid, exception, file=sys.stderr)
            return 0
    else:
        return analyzePDBID(pdbid)

def testPDBIDLoad(pdbid):
    """Returns whether the pdbid downloads and loads correctly.

    :param pdbid:
    :type pdbid: :py:class:`str`

    :return: bool
    :rtype: :py:class:`bool`
    """
    analyzer = densityAnalysis.fromPDBid(pdbid)
    return False if not analyzer else True

def analyzePDBID(pdbid):
    """Process function to analyze a single pdb entry.

    :param pdbid: pdbid for entry to download and analyze.
    :type pdbid: :py:class:`str`

    :return: resultFilename
    :rtype: :py:class:`str`
    """
    startTime = time.process_time()

    analyzer = densityAnalysis.fromPDBid(pdbid)
    if not analyzer or not analyzer.densityElectronRatio:
        return 0

    diffs = {atomType:((analyzer.medians['corrected_density_electron_ratio'][atomType] - analyzer.densityElectronRatio) / analyzer.densityElectronRatio)
              if atomType in analyzer.medians['corrected_density_electron_ratio'] else 0 for atomType in sorted(densityAnalysis.paramsGlobal["radii"])}

    atomOverlapCompleteness = sum(analyzer.atomTypeOverlapCompleteness.values())
    atomOverlapInCompleteness = sum(analyzer.atomTypeOverlapIncompleteness.values())
    if atomOverlapCompleteness > 0 or atomOverlapInCompleteness > 0:
        atomOverlapCompleteness = atomOverlapCompleteness / (atomOverlapCompleteness + atomOverlapInCompleteness)

    stats = {'density_electron_ratio' : analyzer.densityElectronRatio, 'voxel_volume' : analyzer.densityObj.header.unitVolume, 'f000' : analyzer.F000, 'num_voxels_aggregated' : analyzer.numVoxelsAggregated,
        'total_aggregated_electrons' : analyzer.totalAggregatedElectrons, 'density_mean' : analyzer.densityObj.header.densityMean, 'diff_density_mean' : analyzer.diffDensityObj.header.densityMean,
        'resolution' : analyzer.pdbObj.header.resolution, 'space_group' : analyzer.pdbObj.header.spaceGroup, 'num_atoms_analyzed' : len(analyzer.atomCloudDescriptions),
        'num_residue_clouds_analyzed' : len(analyzer.residueCloudDescriptions), 'num_domain_clouds_analyzed' : len(analyzer.domainCloudDescriptions), 'atom_overlap_completeness' : atomOverlapCompleteness}

    properties = { property : value for (property,value) in analyzer.biopdbObj.header.items() }
    properties['residue_counts'] = dict(collections.Counter(residue.resname for residue in analyzer.biopdbObj.get_residues()))
    properties['element_counts'] = dict(collections.Counter(atom.element for atom in analyzer.biopdbObj.get_atoms()))

    elapsedTime = time.process_time() - startTime
    resultFilename = fileUtils.createTempJSONFile({ "pdbid" : analyzer.pdbid, "diffs" : diffs, "stats" : stats, "execution_time" : elapsedTime, "properties" : properties }, "tempResults_")
    analyzer = 0
    gc.collect()
    return resultFilename


class timeout:
    """Implements a timeout context manager to work with a with statement."""
    def __init__(self, seconds=1, error_message='Timeout'):
        """

        :param seconds:
        :type seconds: :py:class:`int`
        :param error_message:
        :type error_message: :py:class:`str`
        """
        self.seconds = seconds
        self.error_message = error_message
    def handle_timeout(self, signum, frame):
        raise TimeoutError(self.error_message)
    def __enter__(self):
        signal.signal(signal.SIGALRM, self.handle_timeout)
        signal.alarm(self.seconds)
    def __exit__(self, type, value, traceback):
        signal.alarm(0)



from pdb_eda import pdbParser
import urllib.request

pdburlPrefix = "http://ftp.rcsb.org/pub/pdb/data/structures/all/pdb/"
pdbfile = 'pdb1g3v.ent.gz'
url = pdburlPrefix + pdbfile
urllib.request.urlretrieve(url, pdbfile)

with gzip.open(pdbfile, 'rt') as gzipFile:
    pdbObj = pdbParser.readPDBfile(gzipFile)

print(pdbObj.header.pdbid)


#!/usr/bin/python3
"""
pdb_eda command-line interface
    The PDB Electron Density Analysis package has facilities for analyzing PDB entries and their associated 2Fo-Fc and Fo-Fc electron density maps.

Usage:
    pdb_eda -h | --help     for this screen.
    pdb_eda --full-help     help documentation on all modes.
    pdb_eda --version       for the version of pdb_eda.
    pdb_eda single ...      for single structure analysis mode. (Most useful command line mode).
    pdb_eda multiple ...    for multiple structure analysis mode. (Second most useful command line mode).
    pdb_eda contacts ...    for crystal contacts analysis mode.  (Third most useful command line mode).
    pdb_eda generate ...    for generating starting parameters file that then needs to be optimized. (Rarely used mode).
    pdb_eda optimize ...    for parameter optimization mode. (Rarely used mode).

For help on a specific mode, use the mode option -h or --help.
For example:
    pdb_eda single --help   for help documentation about single structure analysis mode.
"""

import sys
from . import singleStructure
from . import multipleStructures
from . import crystalContacts
from . import generateParams
from . import optimizeParams
from . import __version__

def main():
    if len(sys.argv) > 1 and sys.argv[1] == "single":
        singleStructure.main()
    elif len(sys.argv) > 1 and sys.argv[1] == "multiple":
        multipleStructures.main()
    elif len(sys.argv) > 1 and sys.argv[1] == "optimize":
        optimizeParams.main()
    elif len(sys.argv) > 1 and sys.argv[1] == "contacts":
        crystalContacts.main()
    elif len(sys.argv) > 1 and sys.argv[1] == "generate":
        generateParams.main()
    elif len(sys.argv) > 1 and (sys.argv[1] == "--version" or sys.argv[1] == "-v") :
        print("Version: ",__version__)
    elif len(sys.argv) > 1 and sys.argv[1] == "--full-help":
        print(__doc__)
        print("-"*80)
        print(singleStructure.__doc__)
        print("-"*80)
        print(multipleStructures.__doc__)
        print("-"*80)
        print(crystalContacts.__doc__)
        print("-"*80)
        print(generateParams.__doc__)
        print("-"*80)
        print(optimizeParams.__doc__)
    else:
        print(__doc__)


if __name__ == "__main__": # this is hidden from the console script created by pip, since main() is called directly.
    if len(sys.argv) > 1 and sys.argv[1] == "--profile":
        sys.argv.pop(1)
        import cProfile
        profiler = cProfile.Profile()
        profiler.enable()
        main()
        profiler.disable()
        profiler.print_stats()
    else:
        main()


#!/usr/bin/env python3

import os
import sys
import re
from setuptools import setup, find_packages, Extension
import Cython.Build
import numpy

if sys.argv[-1] == 'publish':
    os.system('python3 setup.py sdist')
    os.system('twine upload dist/*')
    sys.exit()

def readme():
    with open('README.rst') as readme_file:
        return readme_file.read()

def find_version():
    with open('pdb_eda/__init__.py', 'r') as fd:
        version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',fd.read(), re.MULTILINE).group(1)
    if not version:
        raise RuntimeError('Cannot find version information')
    return version

REQUIRES = [
    "docopt >= 0.6.2",
    "jsonpickle >= 1.4.1",
    "numpy >= 1.16.2",
    "scipy >= 1.2.1",
    "biopython >= 1.77",
    "PyCifRW >= 4.4.2",
    "cython >= 0.29.21"
]

SETUP_REQUIRES = [
    "numpy >= 1.14.5",
    "cython >= 0.29.21"
]

EXTENSIONS = [
    Extension("pdb_eda.cutils", sources=["pdb_eda/cutils.pyx"], extra_compile_args=['-O3'], include_dirs=[numpy.get_include()])
]

setup(
    name='pdb_eda',
    version=find_version(),
    author='Sen Yao, Hunter N.B. Moseley',
    author_email='hunter.moseley@gmail.com',
    description='Methods for analyzing electron density maps in wwPDB',
    keywords='PDB, electron densiy map',
    license='Modified Clear BSD License',
    url='https://github.com/MoseleyBioinformaticsLab/pdb_eda',
    cmdclass={'build_ext': Cython.Build.build_ext},
    packages=find_packages(),
    package_data={'pdb_eda': ['conf/*.json*']},
    platforms='any',
    long_description=readme(),
    setup_requires=SETUP_REQUIRES,
    install_requires=REQUIRES,
    ext_modules=EXTENSIONS,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    entry_points={"console_scripts": ["pdb_eda = pdb_eda.__main__:main"]},
)
